#include "Common.hpp"
#include "Search.hpp"
#include "TrialDivision.hpp"
#include "MillerRabin.hpp"
using namespace std;

int main(int argc, char *argv[]) {
  string method = "";
  int threads = -1;
  int size = -1;
  long seed = -1;
  FILE *output = NULL;

  for (int i = 1; i < argc; ++i) {
    string arg = argv[i];
    string::size_type index = arg.find('=');
    if (index == string::npos) {
      cerr << "illegal option " << arg << "\n";
      return 1;
    } else {
      string left = arg.substr(0, index);
      string right = arg.substr(index + 1);
      if (left == "method") {
        method = right;
      } else if (left == "threads") {
        stringstream ss(right);
        if (!(ss >> threads) || threads <= 0) {
          cerr << "illegal number of threads: " << right << "\n";
          return 1;
        }
      } else if (left == "size") {
        stringstream ss(right);
        if (!(ss >> size) || size <= 31) {
          cerr << "illegal size: " << right << "\n";
          return 1;
        }
      } else if (left == "seed") {
        stringstream ss(right);
        if (!(ss >> seed) || seed < 0) {
          cerr << "illegal seed: " << right << "\n";
          return 1;
        }
      } else if (left == "output") {
        output = fopen(right.c_str(), "w");
        if (!output) {
          cerr << "illegal output filename: " << right << "\n";
          return 1;
        }
      } else {
        cerr << "illegal option " << left << "\n";
        return 1;
      }
    }
  }
  if (method == "") {
    cerr << "method not specified\n";
    return 1;
  }
  if (size == -1) {
    cerr << "size not specified\n";
  }

  if (threads == -1) {
    threads = 0;
  }
  if (seed == -1) {
    seed = (unsigned long)time(NULL);
  }

  TrialDivision::Init(100000000);
  MillerRabin::Init(128);  
  mpz_t prime;
  mpz_init(prime);

  clock_t start = clock();
  unsigned processed = 0;
  if (method == "simple") {
    processed = SimpleSearch(prime, size, seed);
  } else if (method == "threaded") {
    processed = ThreadedSearch(prime, size, seed, threads);
  } else if (method == "mpi") {
    processed = MpiSearch(prime, size, seed, &argc, &argv);
  } else if (method == "omp") {
    processed = OmpSearch(prime, size, seed);
  } else {
    cerr << "illegal method " << method << "\n";
    return 1;
  }
  double time = (double)(clock() - start) / CLOCKS_PER_SEC;
  if (processed != 0) {
    cout << "processed " << processed << " numbers in ";
    cout << time << " seconds (" << processed / time << " n/s)\n";
    if (output) {
      gmp_fprintf(output, "%Zd", prime);
      fclose(output);
    }
  }
  mpz_clear(prime);
  return 0;
}
