#include "Common.hpp"
#include "MillerRabin.hpp"

unsigned MillerRabin::nTrials_;

void MillerRabin::Init(unsigned nTrials) {
  nTrials_ = nTrials;
}

MillerRabin::MillerRabin() {
  mpz_init(minus1_);
  mpz_init(d_);
  mpz_init(x_);
}

MillerRabin::~MillerRabin() {
  mpz_clear(minus1_);
  mpz_clear(d_);
  mpz_clear(x_);
}

bool MillerRabin::Test(mpz_t number, gmp_randstate_t rand) {
  mpz_sub_ui(minus1_, number, 1);
  s_ = mpz_scan1(minus1_, 0);
  mpz_tdiv_q_2exp(d_, minus1_, s_);

  for (unsigned i = 0; i < nTrials_; ++i) {
    if (!Trial(number, rand)) {
      return false;
    }
  }
  return true;
}

bool MillerRabin::Trial(mpz_t number, gmp_randstate_t rand) {
  mpz_urandomm(x_, rand, number);
  if (mpz_cmp_ui(x_, 2) < 0) {
    mpz_add_ui(x_, x_, 2);
  }

  mpz_powm(x_, x_, d_, number);
  if (mpz_cmp_ui(x_, 1) == 0 || mpz_cmp(x_, minus1_) == 0) {
    return true;
  }
  for (unsigned r = 1; r < s_; ++r) {
    mpz_powm_ui(x_, x_, 2, number);
    if (mpz_cmp_ui(x_, 1) == 0) {
      return false;
    } else if (mpz_cmp(x_, minus1_) == 0) {
      return true;
    }
  }
  return false;
}
