#include "Common.hpp"
#include "Search.hpp"
#include "TrialDivision.hpp"
#include "MillerRabin.hpp"

unsigned OmpSearch(mpz_t rop, unsigned size, unsigned long seed) {
  gmp_randstate_t genRand;
  unsigned processed = 0;
  bool done;

  gmp_randinit_default(genRand);
  gmp_randseed_ui(genRand, seed);

  #pragma omp parallel shared(rop, genRand, processed, done)
  {
    gmp_randstate_t mrRand;
    mpz_t number;
    TrialDivision trial;
    MillerRabin mr;

    gmp_randinit_default(mrRand);
    gmp_randseed_ui(mrRand, seed);
    mpz_init(number);

    while (!done) {
      #pragma omp critical
      {
        RandomNumber(number, genRand, size);
        ++processed;
      }
      if (trial.Test(number) && mr.Test(number, mrRand)) {
        #pragma omp critical
        {
          done = true;
          mpz_swap(rop, number);
        }
      }
    }
    gmp_randclear(mrRand);
    mpz_clear(number);
  }

  gmp_randclear(genRand);
  return processed;
}
