#include "Common.hpp"
#include "ThreadWorker.hpp"
#include "TrialDivision.hpp"
#include "MillerRabin.hpp"

ThreadWorker::ThreadWorker(unsigned size, unsigned long seed, unsigned nThreads)
: Worker(size, seed), seed_(seed), nThreads_(nThreads) {
  mpz_init(result_);
  if (nThreads_ == 0) {
    nThreads_ = boost::thread::hardware_concurrency();
    if (nThreads_ == 0) {
      nThreads_ = 1;
    }
  }
}

ThreadWorker::~ThreadWorker() {
  mpz_clear(result_);
}

void ThreadWorker::Search(mpz_t rop) {
  for (unsigned i = 0; i < nThreads_; ++i) {
    boost::thread *thread = new boost::thread(Thread, this, seed_);
    threads_.add_thread(thread);
  }
  boost::unique_lock<boost::mutex> lock(resultMutex_);
  resultCondition_.wait(lock);
  threads_.interrupt_all();
  mpz_swap(rop, result_);
  lock.unlock();
  threads_.join_all();
}

void ThreadWorker::NextCandidate(mpz_t rop) {
  boost::lock_guard<boost::mutex> lock(candidateMutex_);
  Worker::NextCandidate(rop);
}

void ThreadWorker::Finish(mpz_t prime) {
  boost::lock_guard<boost::mutex> lock(resultMutex_);
  mpz_set(result_, prime);
  resultCondition_.notify_one();
}

void ThreadWorker::Thread(ThreadWorker *worker, unsigned long seed) {
  gmp_randstate_t rand;
  gmp_randinit_default(rand);
  gmp_randseed_ui(rand, seed);
  
  mpz_t number;
  mpz_init(number);

  TrialDivision trial;
  MillerRabin mr;

  while (!boost::this_thread::interruption_requested()) {
    worker->NextCandidate(number);
    if (trial.Test(number) && mr.Test(number, rand)) {
      worker->Finish(number);
    }
  }
  gmp_randclear(rand);
  mpz_clear(number);
}
